#The COPYRIGHT file at the top level of this repository contains the full
#copyright notices and license terms.
from decimal import Decimal
import datetime
from trytond.model import Workflow, ModelView, ModelSQL, fields
from trytond.pyson import Eval, If, Not, Equal
from trytond.transaction import Transaction
from trytond.pool import Pool, PoolMeta


__metaclass__ = PoolMeta
__all__ = ['Statement', 'StatementLine', 'StatementMoveLine']

_STATES = {'readonly': Eval('state') != 'draft'}
_DEPENDS = ['state']
CONFIRMED_STATES = {
    'readonly': Not(Equal(Eval('state'), 'draft'))
    }
CONFIRMED_DEPENDS = ['state']
POSTED_STATES = {
    'readonly': Not(Equal(Eval('state'), 'confirmed'))
    }
POSTED_DEPENDS = ['state']
_ZERO = Decimal("0.0")


class Statement(Workflow, ModelSQL, ModelView):
    'Bank Statement'
    __name__ = 'account.bank.statement'
    _rec_name = 'date'

    bank = fields.Many2One('bank', 'Bank', states=_STATES)
    company = fields.Many2One('company.company', 'Company', required=True,
        states=_STATES, depends=['state'], select=True, domain=[
            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
                Eval('context', {}).get('company', 0)),
            ])
    date = fields.DateTime('Date', required=True, states=_STATES,
        depends=['state'], help='Created date bank statement')
    account = fields.Many2One('account.account', 'Account',
        required=True, domain=[
            ('company', '=', Eval('company')),
            ('kind', '!=', 'view'),
            ], states=_STATES)
    journal = fields.Many2One('account.journal', 'Journal',
            required=False, states=_STATES)
    period = fields.Many2One('account.period', 'Period',
        domain=[('state', '=', 'open')], states=_STATES)
    start_balance = fields.Numeric('Account Start Balance',
        required=False, digits=(16, 2), depends=['state'], states=_STATES)
    end_balance = fields.Function(fields.Numeric('Account End Balance',
        required=False, digits=(16, 2), states=_STATES,
        depends=['state']), 'get_end_balance')
    remainder = fields.Function(fields.Numeric('Remainder',
        digits=(16, 2), states=_STATES, depends=['state']),
        'get_remainder')
    total_remainder = fields.Function(fields.Numeric('Total Remainder',
        digits=(16, 2), states=_STATES, depends=['state', 'remainder']),
        'get_total_remainder')
    bank_start_balance = fields.Numeric('Bank Start Balance',
        required=True, digits=(16, 2), depends=['state'], states=_STATES)
    bank_end_balance = fields.Numeric('Bank End Balance',
        required=True, digits=(16, 2), depends=['state'], states=_STATES)
    lines = fields.One2Many('account.bank.statement.line', 'statement',
        'Lines', states=_STATES, context={'period': Eval('period')})
    state = fields.Selection([
            ('draft', 'Draft'),
            ('confirmed', 'Confirmed'),
            ('canceled', 'Canceled'),
            ], 'State', required=True, readonly=True)
    lines_no_transaction = fields.Function(fields.Many2Many('account.move.line',
        None, None, 'Moves Lines without Transaction', domain=[
            ('period', '=', Eval('period')),
            ('account', '=', Eval('account')),
        ]), 'get_lines_no_transaction')

    @classmethod
    def __setup__(cls):
        super(Statement, cls).__setup__()
        cls._transitions |= set((
                ('draft', 'confirmed'),
                ('confirmed', 'canceled'),
                ('canceled', 'draft'),
                ('draft', 'canceled'),
                ))
        cls._buttons.update({
                'confirm': {
                    'invisible': ~Eval('state').in_(['draft']),
                    'icon': 'tryton-go-next',
                    },
                'draft': {
                    'invisible': ~Eval('state').in_(['canceled']),
                    'icon': 'tryton-clear',
                    },
                'cancel': {
                    'invisible': Eval('state').in_(['canceled']),
                    'icon': 'tryton-cancel',
                    },
                })
        cls._error_messages.update({
                'cannot_delete': ('Statement "%s" cannot be deleted because it '
                    'contains lines.'),
                })

    @staticmethod
    def default_company():
        return Transaction().context.get('company')

    @staticmethod
    def default_state():
        return 'draft'

    @staticmethod
    def default_date():
        return datetime.datetime.now()

    @staticmethod
    def default_start_balance():
        return Decimal('0.0')

    @classmethod
    @ModelView.button
    @Workflow.transition('confirmed')
    def confirm(cls, statements):
        StatementLine = Pool().get('account.bank.statement.line')
        lines = []
        for statement in statements:
            lines += statement.lines
        StatementLine.confirm(lines)

    @classmethod
    @ModelView.button
    @Workflow.transition('canceled')
    def cancel(cls, statements):
        StatementLine = Pool().get('account.bank.statement.line')
        lines = []
        for statement in statements:
            for line in statement.lines:
                if not line.state in ('draft', 'canceled'):
                    line.raise_user_error('cannot_cancel_statement', {
                        'line': line.rec_name,
                        'state': line.state,
                    })
            lines += statement.lines
        StatementLine.cancel(lines)

    @classmethod
    @ModelView.button
    @Workflow.transition('draft')
    def draft(cls, statements):
        StatementLine = Pool().get('account.bank.statement.line')
        lines = []
        for statement in statements:
            for line in statement.lines:
                if not line.state in ('canceled', 'draft'):
                    raise "ups"
            lines += statement.lines
        StatementLine.cancel(lines)

    @classmethod
    def delete(cls, statements):
        for statement in statements:
            if statement.lines:
                cls.raise_user_error('cannot_delete', statement.rec_name)
        super(Statement, cls).delete(statements)

    def get_remainder(self, name):
        return (self.bank_end_balance - self.end_balance)

    def get_total_remainder(self, name):
        val = sum([l.debit - l.credit for l in self.lines_no_transaction])
        return (val + self.remainder)

    def get_end_balance(self, name):
        balance = self.start_balance or Decimal(0)
        MoveLine = Pool().get('account.move.line')
        move_lines = MoveLine.search([
                ('period', '=', self.period.id),
                ('account', '=', self.account.id),
            ])
        balance += sum([l.debit - l.credit for l in move_lines])
        return balance

    def get_bank_end_balance(self, name):
        balance = self.bank_start_balance or Decimal(0)
        balance += sum([line.amount for line in self.lines])
        return balance

    def get_lines_no_transaction(self, name):
        if self.period and self.account:
            MoveLine = Pool().get('account.move.line')
            lines = MoveLine.search([
                ('period', '=', self.period.id),
                ('account', '=', self.account.id),
            ])
            targets = []
            for line in lines:
                if not line.bank_lines:
                    targets.append(line.id)
                else:
                    for bl in line.bank_lines:
                        if not bl.statement_line or \
                            bl.statement_line.period != self.period:
                            targets.append(line.id)
            return targets 


class StatementLine(Workflow, ModelSQL, ModelView):
    'Bank Statement Line'
    __name__ = 'account.bank.statement.line'
    _order = [
        ('date', 'ASC'),
        ('sequence', 'ASC')
        ]
    _rec_name = 'description'
    statement = fields.Many2One('account.bank.statement', 'Statement',
        required=True, states=CONFIRMED_STATES)
    date = fields.DateTime('Date', required=True,
        states=CONFIRMED_STATES)
    sequence = fields.Integer('Sequence')
    description = fields.Char('Description', required=True,
        states=CONFIRMED_STATES)
    amount = fields.Numeric('Amount', depends=['currency_digits'],
        digits=(16, 2), required=True, states=CONFIRMED_STATES)
    state = fields.Selection([
            ('draft', 'Draft'),
            ('confirmed', 'Confirmed'),
            ('canceled', 'Canceled'),
            ('posted', 'Posted'),
            ], 'State', required=True, readonly=True)
    bank_lines = fields.One2Many('account.bank.reconciliation',
        'statement_line', 'Bank Lines', states=POSTED_STATES,
            depends=POSTED_DEPENDS, domain=[
                ('statement_line', 'in', (None, Eval('id'))),
                ('move_line.account', 'in', [
                    Eval('credit_account'),
                    Eval('debit_account'),
                ])
            ])
    reconciled = fields.Function(fields.Boolean('Reconciled'),
        'get_reconciled')
    moves_amount = fields.Function(fields.Numeric('Moves Amount',
            digits=(16, 2), depends=['amount']), 'get_moves_amount')
    #Change name to transaction_lines
    lines = fields.One2Many('account.bank.statement.move.line',
        'line', 'Transactions', states=POSTED_STATES,
        context={
            'amount': Eval('amount'),
            'date': Eval('date'),
            'moves_amount': Eval('moves_amount'),
            'description': Eval('description'),
            })
    company = fields.Many2One('company.company', 'Company', required=False,
        domain=[
            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
                Eval('context', {}).get('company', 0)),
            ], select=True, states=CONFIRMED_STATES)
    period = fields.Function(fields.Many2One('account.period',
            'Period'), 'get_period')
    journal = fields.Function(fields.Many2One('account.journal',
            'Journal'), 'get_journal', searcher='search_journal')
    credit_account = fields.Function(fields.Many2One('account.account',
            'Credit Account'), 'get_accounts')
    debit_account = fields.Function(fields.Many2One('account.account',
            'Debit Account'), 'get_accounts')

    @classmethod
    def __setup__(cls):
        super(StatementLine, cls).__setup__()
        cls._transitions |= set((
                ('draft', 'confirmed'),
                ('confirmed', 'posted'),
                ('canceled', 'draft'),
                ('draft', 'canceled'),
                ('confirmed', 'canceled'),
                ('posted', 'canceled'),
                ))
        cls._buttons.update({
                'confirm': {
                    'invisible': ~Eval('state').in_(['draft']),
                    'icon': 'tryton-go-next',
                    },
                'post': {
                    'invisible': Eval('state').in_([
                            'draft', 'canceled', 'posted']),
                    'icon': 'tryton-ok',
                    },
                'cancel': {
                    'invisible': Eval('state').in_(['canceled']),
                    'icon': 'tryton-cancel',
                    },
                'draft': {
                    'invisible': ~Eval('state').in_(['canceled']),
                    'icon': 'tryton-clear',
                    },
                'search_reconcile': {
                    'invisible': ~Eval('state').in_(['confirmed']),
                    'icon': 'tryton-find',
                    },
                })
        cls._error_messages.update({
                'different_amounts': ('Moves Amount "%(moves_amount)s" does '
                    'not match Amount "%(amount)s" in line "%(line)s".'),
                'cannot_cancel_statement_line': ('Line "%(line)s" cannot be '
                    'cancelled.'),
                'cannot_delete': ('Line "%s" cannot be deleted because '
                    'it is not in Draft or Cancelled state.'),
                })

    def _search_bank_line_reconciliation(self):
        BankLines = Pool().get('account.bank.reconciliation')
        lines = BankLines.search([
            ('amount', '=', self.amount),
            ('statement_line', '=', None),
            ])
        if len(lines) == 1:
            line, = lines
            line.statement_line = self.id
            line.save()

    def _search_reconciliation(self):
        self._search_bank_line_reconciliation()

    def get_moves_amount(self, name):
        res1 = sum([l.amount for l in self.bank_lines if l.amount])
        return res1

    @classmethod
    @ModelView.button
    def search_reconcile(cls, st_lines):
        for st_line in st_lines:
            st_line._search_reconciliation()

    @staticmethod
    def default_company():
        return Transaction().context.get('company')

    @staticmethod
    def default_state():
        return 'draft'

    def get_reconciled(self, name):
        return (self.amount == self.moves_amount)

    @classmethod
    def get_accounts(cls, lines, names):
        res = {}
        for name in names:
            res[name] = {}

        for line in lines:
            journal = line.statement.journal
            for name in names:
                account = getattr(journal, name)
                res[name][line.id] = account and account.id
        return res

    @classmethod
    @ModelView.button
    @Workflow.transition('confirmed')
    def confirm(cls, lines):
        pass

    @classmethod
    @ModelView.button
    @Workflow.transition('posted')
    def post(cls, st_lines):
        for st_line in st_lines:
            for line in st_line.lines:
                line.create_move()

    @classmethod
    @ModelView.button
    @Workflow.transition('draft')
    def draft(cls, lines):
        pass

    @classmethod
    @ModelView.button
    @Workflow.transition('canceled')
    def cancel(cls, lines):
        Line = Pool().get('account.bank.reconciliation')

        unlink = []
        for line in lines:
            unlink += line.bank_lines
        Line.write(unlink, {
                'statement_line': None,
                })
        cls.write(lines, {
            'state': 'canceled',
            })

    def get_journal(self, name):
        return self.statement.journal.id

    def get_period(self, name):
        return self.statement.period.id

    @classmethod
    def validate(cls, lines):
        for line in lines:
            line.check_amounts()

    def check_amounts(self):
        if self.state == 'posted' and self.moves_amount != self.amount:
            self.raise_user_error('different_amounts', {
                    'moves_amount': self.moves_amount,
                    'amount': self.amount,
                    'line': self.rec_name,
                    })

    @classmethod
    def search(cls, args, offset=0, limit=None, order=None, count=False,
            query=False):
        """
        Override default search function so that if the user sorts by one field
        only, then 'sequence' is always added as a second sort field. This is
        specially important when the user sorts by date (the most usual) to
        ensure all moves from the same date are properly sorted.
        """
        if order is None:
            order = []
        order = list(order)
        if len(order) == 1:
            order.append(('sequence', order[0][1]))
        return super(StatementLine, cls).search(args, offset, limit, order,
            count, query)

    @classmethod
    def delete(cls, lines):
        for line in lines:
            if line.state not in ('draft', 'canceled'):
                cls.raise_user_error('cannot_delete', line.rec_name)
        super(StatementLine, cls).delete(lines)

    def create_move(self, line):
        pool = Pool()
        Move = pool.get('account.move')
        Line = pool.get('account.move.line')
        Period = pool.get('account.period')

        if line.reconciliation:
            return

        # Create Move
        period_id = Period.find(self.company, date=self.account_date.date())
        move_lines = self._get_counterpart_move_lines(line)
        move = Move(
            period=period_id,
            journal=self.journal.journal,
            lines=move_lines,
            date=self.account_date.date())
        move.save()
        Move.post([move])

        journal = self.journal
        accounts = [journal.credit_account, journal.debit_account]

        # Reconcile lines
        counterpart, = [x for x in move.lines if x.account not in accounts]
        Line.reconcile([counterpart, line])

        # Assign line to  Transactions
        st_move_line, = [x for x in move.lines if x.account in accounts]
        bank_line, = st_move_line.bank_lines
        bank_line.statement_line = self
        bank_line.save()
        self.save()


class StatementMoveLine(ModelSQL, ModelView):
    'Statement Move Line'
    __name__ = 'account.bank.statement.move.line'
    line = fields.Many2One('account.bank.statement.line', 'Line',
        required=True, ondelete='CASCADE')
    date = fields.Date('Date', required=True)
    amount = fields.Numeric('Amount', required=True, digits=(16, 2))
    party = fields.Many2One('party.party', 'Party')
    account = fields.Many2One('account.account', 'Account', required=True,
        domain=[
            ('company', '=', Eval('_parent_line', {}).get('company', 0)),
            ('kind', '!=', 'view'),
            ])
    description = fields.Char('Description')
    move = fields.Many2One('account.move', 'Account Move', readonly=True)

    @classmethod
    def __setup__(cls):
        super(StatementMoveLine, cls).__setup__()
        cls._sql_constraints += [(
                'check_bank_move_amount', 'CHECK(amount != 0)',
                'Amount should be a positive or negative value.'),
        ]
        cls._error_messages.update({
                'debit_credit_account_not_bank_reconcile': (
                    'The credit or debit account of Journal "%s" is not '
                    'checked as "Bank Conciliation".'),
                'missing_account_journal': ('Missing account '
                    'credit or debit on journal "%s".'),
                'same_debit_credit_account': ('Account "%(account)s" in '
                    'statement line "%(line)s" is the same as the one '
                    'configured as credit or debit on journal.'),
                })

    @staticmethod
    def default_amount():
        context = Transaction().context
        if 'amount' in context and 'moves_amount' in context:
            return context['amount'] - context['moves_amount']
        return Decimal(0)

    @staticmethod
    def default_description():
        context = Transaction().context
        if 'description' in context:
            return Transaction().context.get('description')

    @staticmethod
    def default_date():
        if Transaction().context.get('date'):
            return Transaction().context.get('date').date()
        return None

    @fields.depends('account', 'amount', 'party', 'invoice')
    def on_change_party(self):
        res = {}
        if self.party and self.amount:
            if self.amount > Decimal("0.0"):
                account = self.account or self.party.account_receivable
            else:
                account = self.account or self.party.account_payable
            res['account'] = account.id
            res['account.rec_name'] = account.rec_name
        return res

    @fields.depends('amount', 'party', 'account', 'invoice')
    def on_change_amount(self):
        res = {}
        if self.party and not self.account and self.amount:
            if self.amount > Decimal("0.0"):
                account = self.party.account_receivable
            else:
                account = self.party.account_payable
            res['account'] = account.id
            res['account.rec_name'] = account.rec_name
        return res

    def get_rec_name(self, name):
        return self.line.rec_name

    @classmethod
    def copy(cls, lines, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default.setdefault('move', None)
        return super(StatementMoveLine, cls).copy(lines, default=default)

    def create_move(self):
        '''
        Create move for the statement line and return move if created.
        '''
        pool = Pool()
        Move = pool.get('account.move')
        Period = pool.get('account.period')

        if self.move:
            return

        period_id = Period.find(self.line.company.id, date=self.date)

        move_lines = self._get_move_lines()
        move = Move(
            period=period_id,
            journal=self.line.journal,
            date=self.date,
            lines=move_lines,
            description=self.description,
        )
        move.save()
        Move.post([move])

        journal = self.line.journal
        accounts = [journal.credit_account, journal.debit_account]
        st_move_line, = [x for x in move.lines if x.account in accounts]
        bank_line, = st_move_line.bank_lines
        bank_line.statement_line = self.line
        bank_line.save()

        self.move = move
        self.save()
        return move

    @classmethod
    def post_move(cls, lines):
        Move = Pool().get('account.move')
        Move.post([l.move for l in lines if l.move])

    @classmethod
    def delete_move(cls, lines):
        Move = Pool().get('account.move')
        Move.delete([l.move for l in lines if l.move])

    def _get_move_lines(self):
        '''
        Return the move lines for the statement line
        '''
        pool = Pool()
        MoveLine = pool.get('account.move.line')
        amount = self.amount

        move_lines = []
        move_lines.append(MoveLine(
                description=self.description,
                debit=amount < _ZERO and -amount or _ZERO,
                credit=amount >= _ZERO and amount or _ZERO,
                account=self.account,
                party=self.party,
                ))

        journal = self.line.journal
        if self.amount >= _ZERO:
            account = journal.credit_account
        else:
            account = journal.debit_account

        if not account:
            self.raise_user_error('missing_account_journal',
                journal.rec_name)
        if self.account == account:
            self.raise_user_error('same_debit_credit_account', {
                    'account': account.rec_name,
                    'line': self.account.rec_name,
                    })

        bank_move = MoveLine(
            description=self.description,
            debit=amount >= _ZERO and amount or _ZERO,
            credit=amount < _ZERO and -amount or _ZERO,
            account=account,
            party=self.party,
            )
        move_lines.append(bank_move)
        return move_lines
